from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _


class CRUDEvent(models.Model):
    CREATE = 1
    UPDATE = 2
    DELETE = 3
    M2M_CHANGE = 4
    M2M_CHANGE_REV = 5
    M2M_ADD = 6
    M2M_ADD_REV = 7
    M2M_REMOVE = 8
    M2M_REMOVE_REV = 9
    M2M_CLEAR = 10
    M2M_CLEAR_REV = 11

    TYPES = (
        (CREATE, _("Create")),
        (UPDATE, _("Update")),
        (DELETE, _("Delete")),
        (M2M_CHANGE, _("Many-to-Many Change")),
        (M2M_CHANGE_REV, _("Reverse Many-to-Many Change")),
        (M2M_ADD, _("Many-to-Many Add")),
        (M2M_ADD_REV, _("Reverse Many-to-Many Add")),
        (M2M_REMOVE, _("Many-to-Many Remove")),
        (M2M_REMOVE_REV, _("Reverse Many-to-Many Remove")),
        (M2M_CLEAR, _("Many-to-Many Clear")),
        (M2M_CLEAR_REV, _("Reverse Many-to-Many Clear")),
    )

    event_type = models.SmallIntegerField(choices=TYPES, verbose_name=_("Event type"))
    object_id = models.CharField(max_length=255, verbose_name=_("Object ID"))
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        db_constraint=False,
        verbose_name=_("Content type"),
    )
    object_repr = models.TextField(
        default="", blank=True, verbose_name=_("Object representation")
    )
    object_json_repr = models.TextField(
        default="", blank=True, verbose_name=_("Object JSON representation")
    )
    changed_fields = models.TextField(
        default="", blank=True, verbose_name=_("Changed fields")
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        db_constraint=False,
        verbose_name=_("User"),
    )
    user_pk_as_string = models.CharField(
        max_length=255,
        default="",
        blank=True,
        help_text=_("String version of the user pk"),
        verbose_name=_("User PK as string"),
    )
    datetime = models.DateTimeField(auto_now_add=True, verbose_name=_("Date time"))

    class Meta:
        verbose_name = _("CRUD event")
        verbose_name_plural = _("CRUD events")
        ordering = ["-datetime"]
        indexes = [models.Index(fields=["object_id", "content_type"])]

    def is_create(self):
        return self.event_type == self.CREATE

    def is_update(self):
        return self.event_type == self.UPDATE

    def is_delete(self):
        return self.event_type == self.DELETE


class LoginEvent(models.Model):
    LOGIN = 0
    LOGOUT = 1
    FAILED = 2
    TYPES = (
        (LOGIN, _("Login")),
        (LOGOUT, _("Logout")),
        (FAILED, _("Failed login")),
    )
    login_type = models.SmallIntegerField(choices=TYPES, verbose_name=_("Event type"))
    username = models.CharField(
        max_length=255, default="", blank=True, verbose_name=_("Username")
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        db_constraint=False,
        verbose_name=_("User"),
    )
    remote_ip = models.CharField(
        max_length=50, default="", db_index=True, verbose_name=_("Remote IP")
    )
    datetime = models.DateTimeField(auto_now_add=True, verbose_name=_("Date time"))

    class Meta:
        verbose_name = _("login event")
        verbose_name_plural = _("login events")
        ordering = ["-datetime"]


class RequestEvent(models.Model):
    url = models.CharField(null=False, db_index=True, max_length=254, verbose_name=_("URL"))
    method = models.CharField(
        max_length=20, null=False, db_index=True, verbose_name=_("Method")
    )
    query_string = models.TextField(default="", verbose_name=_("Query string"))
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        db_constraint=False,
        verbose_name=_("User"),
    )
    remote_ip = models.CharField(
        max_length=50, default="", db_index=True, verbose_name=_("Remote IP")
    )
    datetime = models.DateTimeField(
        auto_now_add=True, db_index=True, verbose_name=_("Date time")
    )

    class Meta:
        verbose_name = _("request event")
        verbose_name_plural = _("request events")
        ordering = ["-datetime"]



# Generated by Django 4.2.3 on 2023-07-26 18:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('easyaudit', '0017_alter_requestevent_datetime'),
    ]

    operations = [
        migrations.RenameIndex(
            model_name='crudevent',
            new_name='easyaudit_c_object__82020b_idx',
            old_fields=('object_id', 'content_type'),
        ),
    ]



from django.apps import AppConfig


class EasyAuditConfig(AppConfig):
    name = "easyaudit"
    verbose_name = "Easy Audit Application"
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from easyaudit.signals import (  # noqa: F401
            auth_signals,
            model_signals,
            request_signals,
        )



# Generated by Django 2.1 on 2019-07-23 01:26

from django.db import migrations, models
from django.db import connection


def drop_index_if_exists(apps, schema_editor):
    if connection.vendor == 'postgresql':
        schema_editor.execute('DROP INDEX IF EXISTS "easyaudit_requestevent_url_37d1b8c4"')


class Migration(migrations.Migration):

    dependencies = [
        ('easyaudit', '0012_auto_20181018_0012'),
    ]

    operations = [
        migrations.RunPython(drop_index_if_exists, migrations.RunPython.noop),
        migrations.AlterField(
            model_name='requestevent',
            name='url',
            field=models.CharField(db_index=True, max_length=254),
        ),
    ]



# Generated by Django 3.1.1 on 2020-10-19 12:17

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('easyaudit', '0014_auto_20200513_0008'),
    ]

    operations = [
        migrations.AlterField(
            model_name='crudevent',
            name='changed_fields',
            field=models.TextField(blank=True, null=True, verbose_name='Changed fields'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='content_type',
            field=models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='Content type'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date time'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='event_type',
            field=models.SmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Delete'), (4, 'Many-to-Many Change'), (5, 'Reverse Many-to-Many Change'), (6, 'Many-to-Many Add'), (7, 'Reverse Many-to-Many Add'), (8, 'Many-to-Many Remove'), (9, 'Reverse Many-to-Many Remove')], verbose_name='Event type'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_id',
            field=models.CharField(max_length=255, verbose_name='Object ID'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_json_repr',
            field=models.TextField(blank=True, null=True, verbose_name='Object JSON representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_repr',
            field=models.TextField(blank=True, null=True, verbose_name='Object representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user_pk_as_string',
            field=models.CharField(blank=True, help_text='String version of the user pk', max_length=255, null=True, verbose_name='User PK as string'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date time'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='login_type',
            field=models.SmallIntegerField(choices=[(0, 'Login'), (1, 'Logout'), (2, 'Failed login')], verbose_name='Event type'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='remote_ip',
            field=models.CharField(db_index=True, max_length=50, null=True, verbose_name='Remote IP'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='username',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='Username'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date time'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='method',
            field=models.CharField(db_index=True, max_length=20, verbose_name='Method'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='query_string',
            field=models.TextField(null=True, verbose_name='Query string'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='remote_ip',
            field=models.CharField(db_index=True, max_length=50, null=True, verbose_name='Remote IP'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='url',
            field=models.CharField(db_index=True, max_length=254, verbose_name='URL'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
    ]



# Generated by Django 1.11.1 on 2017-06-20 13:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('easyaudit', '0003_auto_20170228_1505'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='crudevent',
            index=models.Index(fields=['object_id', 'content_type'], name='easyaudit_c_object__82020b_idx'),
        ),
    ]



# Generated by Django 5.0.1 on 2024-03-18 19:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('easyaudit', '0018_rename_crudevent_object_id_content_type_index'),
    ]

    operations = [
        migrations.AlterField(
            model_name='crudevent',
            name='changed_fields',
            field=models.TextField(blank=True, default='', verbose_name='Changed fields'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_json_repr',
            field=models.TextField(blank=True, default='', verbose_name='Object JSON representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_repr',
            field=models.TextField(blank=True, default='', verbose_name='Object representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user_pk_as_string',
            field=models.CharField(blank=True, default='', help_text='String version of the user pk', max_length=255, verbose_name='User PK as string'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='remote_ip',
            field=models.CharField(db_index=True, default='', max_length=50, verbose_name='Remote IP'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='username',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Username'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='query_string',
            field=models.TextField(default='', verbose_name='Query string'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='remote_ip',
            field=models.CharField(db_index=True, default='', max_length=50, verbose_name='Remote IP'),
        ),
    ]



# makes easy-audit thread-safe
import contextlib
from typing import Callable

from asgiref.local import Local
from asgiref.sync import iscoroutinefunction, markcoroutinefunction
from django.http.request import HttpRequest
from django.http.response import HttpResponse


class MockRequest:
    def __init__(self, *args, **kwargs):
        user = kwargs.pop("user", None)
        self.user = user
        super().__init__(*args, **kwargs)


_thread_locals = Local()


def get_current_request():
    return getattr(_thread_locals, "request", None)


def get_current_user():
    request = get_current_request()
    if request:
        return getattr(request, "user", None)
    return None


def set_current_user(user):
    try:
        _thread_locals.request.user = user
    except AttributeError:
        request = MockRequest(user=user)
        _thread_locals.request = request


def clear_request():
    with contextlib.suppress(AttributeError):
        del _thread_locals.request


class EasyAuditMiddleware:
    async_capable = True
    sync_capable = True

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        self.get_response = get_response
        if iscoroutinefunction(self.get_response):
            markcoroutinefunction(self)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if iscoroutinefunction(self):
            return self.__acall__(request)

        _thread_locals.request = request
        response = self.get_response(request)

        with contextlib.suppress(AttributeError):
            del _thread_locals.request

        return response

    async def __acall__(self, request: HttpRequest) -> HttpResponse:
        _thread_locals.request = request

        response = await self.get_response(request)

        with contextlib.suppress(AttributeError):
            del _thread_locals.request

        return response



# Generated by Django 3.2.11 on 2022-01-10 16:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('easyaudit', '0015_auto_20201019_1217'),
    ]

    operations = [
        migrations.AlterField(
            model_name='crudevent',
            name='event_type',
            field=models.SmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Delete'), (4, 'Many-to-Many Change'), (5, 'Reverse Many-to-Many Change'), (6, 'Many-to-Many Add'), (7, 'Reverse Many-to-Many Add'), (8, 'Many-to-Many Remove'), (9, 'Reverse Many-to-Many Remove'), (10, 'Many-to-Many Clear'), (11, 'Reverse Many-to-Many Clear')], verbose_name='Event type'),
        ),
    ]



# Generated by Django 4.1.1 on 2022-09-25 00:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('easyaudit', '0016_alter_crudevent_event_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='requestevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Date time'),
        ),
    ]



from django.contrib.auth import get_user_model, signals
from django.db import transaction
from django.utils.module_loading import import_string

from easyaudit.middleware.easyaudit import get_current_request
from easyaudit.models import LoginEvent
from easyaudit.settings import (
    DATABASE_ALIAS,
    LOGGING_BACKEND,
    REMOTE_ADDR_HEADER,
    WATCH_AUTH_EVENTS,
)
from easyaudit.utils import should_propagate_exceptions

audit_logger = import_string(LOGGING_BACKEND)()


def user_logged_in(sender, request, user, **kwargs):
    try:
        with transaction.atomic(using=DATABASE_ALIAS):
            audit_logger.login(
                {
                    "login_type": LoginEvent.LOGIN,
                    "username": getattr(user, user.USERNAME_FIELD),
                    "user_id": getattr(user, "id", None),
                    "remote_ip": request.META.get(REMOTE_ADDR_HEADER, ""),
                }
            )
    except Exception:
        if should_propagate_exceptions():
            raise


def user_logged_out(sender, request, user, **kwargs):
    try:
        with transaction.atomic(using=DATABASE_ALIAS):
            audit_logger.login(
                {
                    "login_type": LoginEvent.LOGOUT,
                    "username": getattr(user, user.USERNAME_FIELD),
                    "user_id": getattr(user, "id", None),
                    "remote_ip": request.META.get(REMOTE_ADDR_HEADER, ""),
                }
            )
    except Exception:
        if should_propagate_exceptions():
            raise


def user_login_failed(sender, credentials, **kwargs):
    try:
        with transaction.atomic(using=DATABASE_ALIAS):
            request = get_current_request()
            user_model = get_user_model()
            audit_logger.login(
                {
                    "login_type": LoginEvent.FAILED,
                    "username": credentials[user_model.USERNAME_FIELD],
                    "remote_ip": request.META.get(REMOTE_ADDR_HEADER, ""),
                }
            )
    except Exception:
        if should_propagate_exceptions():
            raise


if WATCH_AUTH_EVENTS:
    signals.user_logged_in.connect(
        user_logged_in, dispatch_uid="easy_audit_signals_logged_in"
    )
    signals.user_logged_out.connect(
        user_logged_out, dispatch_uid="easy_audit_signals_logged_out"
    )
    signals.user_login_failed.connect(
        user_login_failed, dispatch_uid="easy_audit_signals_login_failed"
    )



# Generated by Django 3.0.6 on 2020-05-13 00:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('easyaudit', '0013_auto_20190723_0126'),
    ]

    operations = [
        migrations.AlterField(
            model_name='crudevent',
            name='object_id',
            field=models.CharField(max_length=255),
        ),
    ]



# ruff: noqa: PLR0913
import json
import logging
from functools import partial

from django.conf import settings
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.db import transaction
from django.db.models import signals
from django.utils.encoding import force_str

from easyaudit.middleware.easyaudit import get_current_request
from easyaudit.models import CRUDEvent
from easyaudit.settings import (
    CRUD_DIFFERENCE_CALLBACKS,
    REGISTERED_CLASSES,
    UNREGISTERED_CLASSES,
    WATCH_MODEL_EVENTS,
)
from easyaudit.utils import model_delta, should_propagate_exceptions

from .crud_flows import (
    m2m_changed_crud_flow,
    post_delete_crud_flow,
    post_save_crud_flow,
    pre_save_crud_flow,
)

logger = logging.getLogger(__name__)


def should_audit(instance):
    """Return True or False to indicate whether the instance should be audited."""
    # do not audit any model listed in UNREGISTERED_CLASSES
    for unregistered_class in UNREGISTERED_CLASSES:
        if isinstance(instance, unregistered_class):
            return False

    # only audit models listed in REGISTERED_CLASSES (if it's set)
    if len(REGISTERED_CLASSES) > 0:
        for registered_class in REGISTERED_CLASSES:
            if isinstance(instance, registered_class):
                break
        else:
            return False

    # all good
    return True


def call_callbacks(
    instance, object_json_repr, created, raw, using, update_fields, **kwargs
) -> bool:
    kwargs["request"] = get_current_request()  # Make request available in callbacks

    return all(
        callback(
            instance,
            object_json_repr,
            created,
            raw,
            using,
            update_fields,
            **kwargs,
        )
        for callback in CRUD_DIFFERENCE_CALLBACKS
        if callable(callback)
    )


def handle_signal_exception(signal):
    logger.exception(f"easy audit had a {signal} exception.")

    if should_propagate_exceptions():
        raise


def pre_save(sender, instance, raw, using, update_fields, **kwargs):
    if raw:
        # Return if loading Fixtures
        return None

    try:
        if not should_audit(instance):
            return False

        with transaction.atomic(using=using):
            try:
                object_json_repr = serializers.serialize("json", [instance])
            except Exception:
                # We need a better way for this to work. ManyToMany will fail on
                # pre_save on create
                return None

            # Determine if the instance is a create
            created = instance.pk is None or instance._state.adding

            # created or updated?
            delta = {}
            if not created:
                old_model = sender.objects.get(pk=instance.pk)
                delta = model_delta(old_model, instance)

                if not delta and getattr(
                    settings,
                    "DJANGO_EASY_AUDIT_CRUD_EVENT_NO_CHANGED_FIELDS_SKIP",
                    False,
                ):
                    return False

            # callbacks
            create_crud_event = call_callbacks(
                instance, object_json_repr, created, raw, using, update_fields, **kwargs
            )

            # Create crud event only if all callbacks returned True
            if create_crud_event and not created:
                crud_flow = partial(
                    pre_save_crud_flow,
                    instance=instance,
                    object_json_repr=object_json_repr,
                    changed_fields=json.dumps(delta),
                )

                if getattr(settings, "TEST", False):
                    crud_flow()
                else:
                    transaction.on_commit(crud_flow, using=using)
    except Exception:
        handle_signal_exception("pre_save")


def post_save(sender, instance, created, raw, using, update_fields, **kwargs):
    if raw:
        # Return if loading Fixtures
        return None

    try:
        if not should_audit(instance):
            return False

        with transaction.atomic(using=using):
            object_json_repr = serializers.serialize("json", [instance])

            # callbacks
            create_crud_event = call_callbacks(
                instance, object_json_repr, created, raw, using, update_fields, **kwargs
            )

            # Create crud event only if all callbacks returned True
            if create_crud_event and created:
                crud_flow = partial(
                    post_save_crud_flow,
                    instance=instance,
                    object_json_repr=object_json_repr,
                )
                if getattr(settings, "TEST", False):
                    crud_flow()
                else:
                    transaction.on_commit(crud_flow, using=using)
    except Exception:
        handle_signal_exception("post_save")


def _m2m_rev_field_name(model1, model2):
    """Get the name of the reverse m2m accessor from `model1` to `model2`.

    For example, if User has a ManyToManyField connected to Group,
    `_m2m_rev_field_name(Group, User)` retrieves the name of the field on
    Group that lists a group's Users. (By default, this field is called
    `user_set`, but the name can be overridden).
    """
    m2m_field_names = [
        rel.get_accessor_name()
        for rel in model1._meta.get_fields()
        if rel.many_to_many and rel.auto_created and rel.related_model == model2
    ]
    return m2m_field_names[0]


def m2m_changed(sender, instance, action, reverse, model, pk_set, using, **kwargs):
    try:
        if not should_audit(instance):
            return False
        if action not in ("post_add", "post_remove", "post_clear"):
            return False

        with transaction.atomic(using=using):
            object_json_repr = serializers.serialize("json", [instance])

            if reverse:
                reverse_actions = {
                    "post_add": CRUDEvent.M2M_ADD_REV,
                    "post_remove": CRUDEvent.M2M_REMOVE_REV,
                    "post_clear": CRUDEvent.M2M_CLEAR_REV,
                }
                event_type = reverse_actions.get(action, CRUDEvent.M2M_CHANGE_REV)

                # Add reverse M2M changes to event. Must use json lib because
                # django serializers ignore extra fields.
                tmp_repr = json.loads(object_json_repr)

                m2m_rev_field = _m2m_rev_field_name(instance._meta.concrete_model, model)
                related_instances = getattr(instance, m2m_rev_field).all()
                related_ids = [r.pk for r in related_instances]

                tmp_repr[0]["m2m_rev_model"] = force_str(model._meta)
                tmp_repr[0]["m2m_rev_pks"] = related_ids
                tmp_repr[0]["m2m_rev_action"] = action
                object_json_repr = json.dumps(tmp_repr, cls=DjangoJSONEncoder)
            else:
                forward_actions = {
                    "post_add": CRUDEvent.M2M_ADD,
                    "post_remove": CRUDEvent.M2M_REMOVE,
                    "post_clear": CRUDEvent.M2M_CLEAR,
                }
                event_type = forward_actions.get(action, CRUDEvent.M2M_CHANGE)

            crud_flow = partial(
                m2m_changed_crud_flow,
                action=action,
                model=model,
                instance=instance,
                pk_set=pk_set,
                event_type=event_type,
                object_json_repr=object_json_repr,
            )
            if getattr(settings, "TEST", False):
                crud_flow()
            else:
                transaction.on_commit(crud_flow, using=using)
    except Exception:
        handle_signal_exception("m2m-changed")


def post_delete(sender, instance, using, **kwargs):
    try:
        if not should_audit(instance):
            return False

        with transaction.atomic(using=using):
            object_json_repr = serializers.serialize("json", [instance])
            # instance.pk returns None if the changes are performed within a transaction
            object_id = instance.pk

            crud_flow = partial(
                post_delete_crud_flow,
                instance=instance,
                object_id=object_id,
                object_json_repr=object_json_repr,
            )
            if getattr(settings, "TEST", False):
                crud_flow()
            else:
                transaction.on_commit(
                    crud_flow,
                    using=using,
                )
    except Exception:
        handle_signal_exception("post-delete")


if WATCH_MODEL_EVENTS:
    signals.post_save.connect(post_save, dispatch_uid="easy_audit_signals_post_save")
    signals.pre_save.connect(pre_save, dispatch_uid="easy_audit_signals_pre_save")
    signals.m2m_changed.connect(m2m_changed, dispatch_uid="easy_audit_signals_m2m_changed")
    signals.post_delete.connect(post_delete, dispatch_uid="easy_audit_signals_post_delete")



# Generated by Django 4.2.11 on 2024-03-20 18:56

from django.conf import settings
from django.db import migrations, models
import django.db.migrations.operations.special
import django.db.models.deletion


# Functions from the following migrations need manual copying.
# Move them and any dependencies into this file, then update the
# RunPython operations to refer to the local versions:
# easyaudit.migrations.0013_auto_20190723_0126

class Migration(migrations.Migration):

    replaces = [('easyaudit', '0004_auto_20170620_1354'), ('easyaudit', '0005_auto_20170713_1155'), ('easyaudit', '0006_auto_20171018_1242'), ('easyaudit', '0007_auto_20180105_0838'), ('easyaudit', '0008_auto_20180220_1908'), ('easyaudit', '0009_auto_20180314_2225'), ('easyaudit', '0010_repr_text'), ('easyaudit', '0011_auto_20181101_1339'), ('easyaudit', '0012_auto_20181018_0012'), ('easyaudit', '0013_auto_20190723_0126'), ('easyaudit', '0014_auto_20200513_0008'), ('easyaudit', '0015_auto_20201019_1217'), ('easyaudit', '0016_alter_crudevent_event_type'), ('easyaudit', '0017_alter_requestevent_datetime'), ('easyaudit', '0018_rename_crudevent_object_id_content_type_index'), ('easyaudit', '0019_alter_crudevent_changed_fields_and_more')]

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('easyaudit', '0003_auto_20170228_1505'),
    ]

    operations = [
        # replaces index_together on object_id and content_type
        migrations.AddIndex(
            model_name='crudevent',
            index=models.Index(fields=['object_id', 'content_type'], name='easyaudit_c_object__82020b_idx'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='event_type',
            field=models.SmallIntegerField(choices=[(1, b'Create'), (2, b'Update'), (3, b'Delete'), (4, b'Many-to-Many Change'), (5, b'Reverse Many-to-Many Change')]),
        ),
        migrations.AddField(
            model_name='loginevent',
            name='remote_ip',
            field=models.CharField(db_index=True, max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='event_type',
            field=models.SmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Delete'), (4, 'Many-to-Many Change'), (5, 'Reverse Many-to-Many Change')]),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user_pk_as_string',
            field=models.CharField(blank=True, help_text='String version of the user pk', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='login_type',
            field=models.SmallIntegerField(choices=[(0, 'Login'), (1, 'Logout'), (2, 'Failed login')]),
        ),
        migrations.AddField(
            model_name='crudevent',
            name='changed_fields',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user_pk_as_string',
            field=models.CharField(blank=True, help_text='String version of the user pk', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_repr',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='content_type',
            field=models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
        migrations.CreateModel(
            name='RequestEvent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.TextField()),
                ('method', models.CharField(db_index=True, max_length=20)),
                ('query_string', models.TextField(null=True)),
                ('remote_ip', models.CharField(db_index=True, max_length=50, null=True)),
                ('datetime', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-datetime'],
                'verbose_name': 'request event',
                'verbose_name_plural': 'request events',
            },
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='url',
            field=models.CharField(db_index=True, max_length=254),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_id',
            field=models.CharField(max_length=255),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='changed_fields',
            field=models.TextField(blank=True, null=True, verbose_name='Changed fields'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='content_type',
            field=models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='Content type'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date time'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='event_type',
            field=models.SmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Delete'), (4, 'Many-to-Many Change'), (5, 'Reverse Many-to-Many Change'), (6, 'Many-to-Many Add'), (7, 'Reverse Many-to-Many Add'), (8, 'Many-to-Many Remove'), (9, 'Reverse Many-to-Many Remove')], verbose_name='Event type'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_id',
            field=models.CharField(max_length=255, verbose_name='Object ID'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_json_repr',
            field=models.TextField(blank=True, null=True, verbose_name='Object JSON representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_repr',
            field=models.TextField(blank=True, null=True, verbose_name='Object representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user_pk_as_string',
            field=models.CharField(blank=True, help_text='String version of the user pk', max_length=255, null=True, verbose_name='User PK as string'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date time'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='login_type',
            field=models.SmallIntegerField(choices=[(0, 'Login'), (1, 'Logout'), (2, 'Failed login')], verbose_name='Event type'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='remote_ip',
            field=models.CharField(db_index=True, max_length=50, null=True, verbose_name='Remote IP'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='username',
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name='Username'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date time'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='method',
            field=models.CharField(db_index=True, max_length=20, verbose_name='Method'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='query_string',
            field=models.TextField(null=True, verbose_name='Query string'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='remote_ip',
            field=models.CharField(db_index=True, max_length=50, null=True, verbose_name='Remote IP'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='url',
            field=models.CharField(db_index=True, max_length=254, verbose_name='URL'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='user',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='event_type',
            field=models.SmallIntegerField(choices=[(1, 'Create'), (2, 'Update'), (3, 'Delete'), (4, 'Many-to-Many Change'), (5, 'Reverse Many-to-Many Change'), (6, 'Many-to-Many Add'), (7, 'Reverse Many-to-Many Add'), (8, 'Many-to-Many Remove'), (9, 'Reverse Many-to-Many Remove'), (10, 'Many-to-Many Clear'), (11, 'Reverse Many-to-Many Clear')], verbose_name='Event type'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='datetime',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Date time'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='changed_fields',
            field=models.TextField(blank=True, default='', verbose_name='Changed fields'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_json_repr',
            field=models.TextField(blank=True, default='', verbose_name='Object JSON representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='object_repr',
            field=models.TextField(blank=True, default='', verbose_name='Object representation'),
        ),
        migrations.AlterField(
            model_name='crudevent',
            name='user_pk_as_string',
            field=models.CharField(blank=True, default='', help_text='String version of the user pk', max_length=255, verbose_name='User PK as string'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='remote_ip',
            field=models.CharField(db_index=True, default='', max_length=50, verbose_name='Remote IP'),
        ),
        migrations.AlterField(
            model_name='loginevent',
            name='username',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Username'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='query_string',
            field=models.TextField(default='', verbose_name='Query string'),
        ),
        migrations.AlterField(
            model_name='requestevent',
            name='remote_ip',
            field=models.CharField(db_index=True, default='', max_length=50, verbose_name='Remote IP'),
        ),
    ]



import re
from importlib import import_module

from django.conf import settings
from django.contrib.auth import SESSION_KEY as AUTH_SESSION_KEY
from django.contrib.auth import get_user_model
from django.contrib.sessions.models import Session
from django.core.signals import request_started
from django.http.cookie import SimpleCookie
from django.utils import timezone
from django.utils.module_loading import import_string

from easyaudit.settings import (
    LOGGING_BACKEND,
    REGISTERED_URLS,
    REMOTE_ADDR_HEADER,
    UNREGISTERED_URLS,
    WATCH_REQUEST_EVENTS,
)

session_engine = import_module(settings.SESSION_ENGINE)
audit_logger = import_string(LOGGING_BACKEND)()


def should_log_url(url):
    # check if current url is blacklisted
    for unregistered_url in UNREGISTERED_URLS:
        pattern = re.compile(unregistered_url)
        if pattern.match(url):
            return False

    # only audit URLs listed in REGISTERED_URLS (if it's set)
    if len(REGISTERED_URLS) > 0:
        for registered_url in REGISTERED_URLS:
            pattern = re.compile(registered_url)
            if pattern.match(url):
                return True
        return False

    # all good
    return True


def request_started_handler(sender, **kwargs):
    environ = kwargs.get("environ")
    scope = kwargs.get("scope")
    if environ:
        path = environ["PATH_INFO"]
        cookie_string = environ.get("HTTP_COOKIE")
        remote_ip = environ.get(REMOTE_ADDR_HEADER, None)
        method = environ["REQUEST_METHOD"]
        query_string = environ["QUERY_STRING"]

    else:
        method = scope.get("method")
        path = scope.get("path")
        headers = dict(scope.get("headers"))
        cookie_string = headers.get(b"cookie")
        if isinstance(cookie_string, bytes):
            cookie_string = cookie_string.decode("utf-8")
        remote_ip = next(iter(scope.get("client", ("0.0.0.0", 0))))  # noqa: S104
        query_string = scope.get("query_string")

    if not should_log_url(path):
        return

    user = None
    # get the user from cookies
    if not user and cookie_string:
        cookie = SimpleCookie()
        cookie.load(cookie_string)
        session_cookie_name = settings.SESSION_COOKIE_NAME
        if session_cookie_name in cookie:
            session_id = cookie[session_cookie_name].value

            try:
                session = session_engine.SessionStore(session_key=session_id).load()
            except Session.DoesNotExist:
                session = None

            if session and AUTH_SESSION_KEY in session:
                user_id = session.get(AUTH_SESSION_KEY)
                try:
                    user = get_user_model().objects.get(id=user_id)
                except Exception:
                    user = None

    # may want to wrap this in an atomic transaction later
    audit_logger.request(
        {
            "url": path,
            "method": method,
            "query_string": query_string,
            "user_id": getattr(user, "id", None),
            "remote_ip": remote_ip,
            "datetime": timezone.now(),
        }
    )


if WATCH_REQUEST_EVENTS:
    request_started.connect(
        request_started_handler, dispatch_uid="easy_audit_signals_request_started"
    )



import contextlib
import json
import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import transaction
from django.utils import timezone
from django.utils.module_loading import import_string

from easyaudit.middleware.easyaudit import get_current_user
from easyaudit.models import CRUDEvent
from easyaudit.settings import DATABASE_ALIAS, LOGGING_BACKEND
from easyaudit.utils import get_m2m_field_name, should_propagate_exceptions

logger = logging.getLogger(__name__)
audit_logger = import_string(LOGGING_BACKEND)()


def get_current_user_details():
    user_id = ""
    user_pk_as_string = ""

    with contextlib.suppress(Exception):
        user = get_current_user()
        if user and not isinstance(user, AnonymousUser):
            if getattr(settings, "DJANGO_EASY_AUDIT_CHECK_IF_REQUEST_USER_EXISTS", True):
                # validate that the user still exists
                user = get_user_model().objects.get(pk=user.pk)
            user_id, user_pk_as_string = user.id, str(user.pk)

    return user_id, user_pk_as_string


def log_event(event_type, instance, object_id, object_json_repr, **kwargs):
    user_id, user_pk_as_string = get_current_user_details()
    with transaction.atomic(using=DATABASE_ALIAS):
        audit_logger.crud(
            {
                "content_type_id": ContentType.objects.get_for_model(instance).id,
                "datetime": timezone.now(),
                "event_type": event_type,
                "object_id": object_id,
                "object_json_repr": object_json_repr or "",
                "object_repr": str(instance),
                "user_id": user_id,
                "user_pk_as_string": user_pk_as_string,
                **kwargs,
            }
        )


def handle_flow_exception(instance, signal):
    instance_str = ""
    with contextlib.suppress(Exception):
        instance_str = f" instance: {instance}, instance pk: {instance.pk}"

    logger.exception(
        f"easy audit had a {signal} exception on CRUDEvent creation.{instance_str}"
    )
    if should_propagate_exceptions():
        raise


def pre_save_crud_flow(instance, object_json_repr, changed_fields):
    try:
        log_event(
            CRUDEvent.UPDATE,
            instance,
            instance.pk,
            object_json_repr,
            changed_fields=changed_fields,
        )
    except Exception:
        handle_flow_exception(instance, "pre_save")


def post_save_crud_flow(instance, object_json_repr):
    try:
        log_event(
            CRUDEvent.CREATE,
            instance,
            instance.pk,
            object_json_repr,
        )
    except Exception:
        handle_flow_exception(instance, "post_save")


def m2m_changed_crud_flow(  # noqa: PLR0913
    action, model, instance, pk_set, event_type, object_json_repr
):
    try:
        if action == "post_clear":
            changed_fields = []
        else:
            changed_fields = json.dumps(
                {get_m2m_field_name(model, instance): list(pk_set)},
                cls=DjangoJSONEncoder,
            )
        log_event(
            event_type,
            instance,
            instance.pk,
            object_json_repr,
            changed_fields=changed_fields,
        )
    except Exception:
        handle_flow_exception(instance, "m2m_changed")


def post_delete_crud_flow(instance, object_id, object_json_repr):
    try:
        log_event(
            CRUDEvent.DELETE,
            instance,
            object_id,
            object_json_repr,
        )

    except Exception:
        handle_flow_exception(instance, "post_delete")
